/******************************************************************************
*
* Copyright (C) 2015 Xilinx, Inc.  All rights reserved.
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* Use of the Software is limited solely to applications:
* (a) running on a Xilinx device, or
* (b) that interact with a Xilinx device through a bus or interconnect.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* XILINX CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* Except as contained in this notice, the name of the Xilinx shall not be used
* in advertising or otherwise to promote the sale, use or other dealings in
* this Software without prior written authorization from Xilinx.
*
******************************************************************************/

/*
 * CONTENT: release_all_r5
 * This application performs the minimum power management setup for the RPU.
 * It calls PmInitFinalize to release all resources it's been assigned based
 * on the PM configuration object passed by the FSBL. Only the TCM memories
 * remain requested.
 * Finally the application performs a self suspend, allowing the RPU power
 * island to be turned off while keeping the TCM memories in retention.
 *
 * Note: if RPU is configured for split mode, then this application must be
 * run for both RPU0 and RPU1 in order to release all RPU resources!
 */

#include <xil_exception.h>
#include <xil_printf.h>
#include <xil_io.h>
#include <xstatus.h>
#include <sleep.h>
#include <ipi.h>
#include <pm_api_sys.h>
#include <pm_client.h>
#include <libeswpm/gic_setup.h>
#include <libeswpm/pm_ipi.h>

extern void __attribute__((weak)) *_vector_table;
static XScuGic GicInst;
static XIpiPsu IpiInst;

/**
 * InitApp() - initialize interrupts and power management
 */
static u32 InitApp(void)
{
	XStatus status;

	// Setting up GIC - not mandatory for minimal application
	status = GicSetupInterruptSystem(&GicInst);
	if (XST_SUCCESS != status) {
		pm_print("Error (%d) in setting up GIC\n", status);
		goto done;
	}

	// Setting up IPIs - only required for receiving callbacks
	status  = PmIpiInit(&GicInst, &IpiInst);
	if (XST_SUCCESS != status) {
		pm_print("Error (%d) in configuring IPI\n", status);
		goto done;
	}

	// Setup xilpm (also not critical if we just want to release resources)
	status = XPm_InitXilpm(&IpiInst);
	if (XST_SUCCESS != status) {
		pm_print("Error (%d) in initializing xilpm\n", status);
		goto done;
	}

	// This releases all resources assigned to the RPU (except TCMs)
	status = XPm_InitFinalize();
	if (XST_SUCCESS != status) {
		pm_print("Failure in call to XPm_InitFinalize()\n");
		goto done;
	}

done:
	return status;
}

int main(void)
{
	XStatus status = InitApp();

	if (XST_SUCCESS != status) {
		return XST_FAILURE;
	}

	pm_print("release_all_r5 init done\n");

	// Initiate suspend procedure
	XPm_SelfSuspend(NODE_RPU, MAX_LATENCY, 0, (u32)&_vector_table);

	// Keep the TCM memories in retention, with no requirements
	XPm_SetRequirement(NODE_TCM_0_A, 0, 0, REQUEST_ACK_NO);
	XPm_SetRequirement(NODE_TCM_0_B, 0, 0, REQUEST_ACK_NO);
	XPm_SetRequirement(NODE_TCM_1_A, 0, 0, REQUEST_ACK_NO);
	XPm_SetRequirement(NODE_TCM_1_B, 0, 0, REQUEST_ACK_NO);

	// Finalize suspend procedure (enter WFI), then RPU will power off
	XPm_SuspendFinalize();

	return XST_FAILURE;
}
